"""
Proxies for libgeos, GEOS-specific exceptions, and utilities
"""
import shapely

geos_version_string = shapely.geos_capi_version_string
geos_version = shapely.geos_version
geos_capi_version = shapely.geos_capi_version
